/*	

Zoulou ! 
							  
(Factor oracle with length repeated suffix).
						
merci : Matthieu Raffinot, Maxime Crochemore, et Cyril Allauzen ;
merci : Arnaud Lefebvre et Thierry Lecroq ;

Version / 13 juin 2010.	
	
*/

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

#include "ext.h"							
#include "ext_obex.h"
#include "stdlib.h"
#include "math.h"	
#include "time.h"	
#include "ext_common.h"
#include "ext_systhread.h"	

#define MAX_RESULT			256
#define INIT_GRAF_SIZE		1024
#define TAILLE_GRAF_MAX		262144

// ----------------------------------------------------------------------------------------------------------------

typedef struct _arc {
	long		dest ;
	long		val ;
	} t_arc ;
	
typedef struct _noeud {
	long		refer ;
	long		lrs ;
	long		straight_val ;
	t_linklist	*list_arcs ;
	} t_noeud ;

enum {
	DEF_REPEAT = 2,
	DEF_THRESHOLD = 3,
	DEF_STRAIGHT = 20
	} ;

// ----------------------------------------------------------------------------------------------------------------

typedef struct _zoulou {
	t_object			ob ;	
	t_noeud				**graf ;
	t_systhread_mutex   m_mutex ;
	t_atomarray			*slice_result ;
	t_atomarray			*get_result ;
	long				graf_size ;             
	long				graf_pos ;			
	long				graf_index ;
	long				repeat ;
	long				straight ;
	long				threshold ;
	long				slice ;
	long				temp ;
	long				flag ;
	long				lock ;
	void				*m_clock ;
	void				*m_out1 ;
	void				*m_out2 ;
	void				*m_out3 ;
	void				*m_out4 ;
	} t_zoulou ;

// ----------------------------------------------------------------------------------------------------------------	

void	*zoulou_new			(t_symbol *s, long argc, t_atom *argv) ;
void	zoulou_bang			(t_zoulou *x) ;
void	zoulou_int			(t_zoulou *x, long n) ;
void	zoulou_clear		(t_zoulou *x) ;
void	zoulou_go			(t_zoulou *x, long n) ;
void	zoulou_get			(t_zoulou *x) ;

void	graf_add			(t_zoulou *x, long n) ;

t_noeud	*noeud_new			(t_noeud *ptr) ;
t_arc	*arc_new			(long d, long v) ;
long	graf_exist			(t_noeud **p, long n, long v) ;       
long	arc_exist			(t_arc *aptr, long *v) ;
long	repeated_suffix		(t_noeud **p, long a, long b) ;  
long	common_suffix		(t_noeud **p, long a, long b) ;

long	graf_step			(t_zoulou *x) ;
void	graf_clear			(t_zoulou *x) ;

void	fill_slice_result	(t_zoulou *x, long ac, long av) ;

void	zoulou_dump			(t_zoulou *x, long n) ;
void	zoulou_task			(t_zoulou *x) ;
void	zoulou_assist		(t_zoulou *x, void *b, long m, long a, char *s) ;
void	zoulou_free			(t_zoulou *x) ;

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

static t_class	*zoulou_class ;

static t_symbol	*my_msg_straight	= NULL ;
static t_symbol	*my_msg_link		= NULL ;
static t_symbol	*my_msg_lrs			= NULL ;

int main (void)
	{	
		t_class *c = NULL ;
		
		my_msg_straight	= gensym ("straight") ;
		my_msg_link		= gensym ("arc") ;
		my_msg_lrs		= gensym ("lrs") ;
		
		c = class_new	("zoulou", (method)zoulou_new, (method)zoulou_free, (long)sizeof(t_zoulou), 0L, A_GIMME, 0) ;

		class_addmethod (c, (method)zoulou_assist,	"assist", A_CANT, 0) ; 
		class_addmethod (c, (method)zoulou_bang,	"bang", 0) ;
		class_addmethod (c, (method)zoulou_int,		"int", A_LONG, 0) ;
		class_addmethod (c, (method)zoulou_clear,	"clear", 0) ;
		class_addmethod (c, (method)zoulou_dump,	"dump_max", A_DEFLONG, 0) ;
		class_addmethod (c, (method)zoulou_go,		"go", A_DEFLONG, 0) ;
		class_addmethod (c, (method)zoulou_get,		"get", 0) ;

		class_register	(CLASS_BOX, c) ; 
		
		CLASS_ATTR_LONG (c, "lrs",			ATTR_FLAGS_NONE, t_zoulou, repeat) ;
		CLASS_ATTR_LONG (c, "straight",		ATTR_FLAGS_NONE, t_zoulou, straight) ;
		CLASS_ATTR_LONG	(c, "threshold",	ATTR_FLAGS_NONE, t_zoulou, threshold) ;
		CLASS_ATTR_LONG	(c, "slice",		ATTR_FLAGS_NONE, t_zoulou, slice) ;
		
		CLASS_ATTR_SAVE (c, "lrs",			ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE (c, "straight",		ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE	(c, "threshold",	ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE	(c, "slice",		ATTR_FLAGS_NONE) ;
		
		CLASS_ATTR_FILTER_CLIP	(c, "lrs",			0, LONG_MAX) ;
		CLASS_ATTR_FILTER_CLIP	(c, "straight",		0, 100) ;
		CLASS_ATTR_FILTER_CLIP	(c, "threshold",	1, MAX_RESULT) ;
		CLASS_ATTR_FILTER_CLIP	(c, "slice",		0, 2) ;
		
		CLASS_ATTR_ENUMINDEX	(c, "slice",		ATTR_FLAGS_NONE, "off maximum threshold") ;
		
		CLASS_ATTR_ORDER (c, "lrs",					ATTR_FLAGS_NONE, "1") ;
		CLASS_ATTR_ORDER (c, "straight",			ATTR_FLAGS_NONE, "2") ;
		CLASS_ATTR_ORDER (c, "slice",				ATTR_FLAGS_NONE, "3") ;
		CLASS_ATTR_ORDER (c, "threshold",			ATTR_FLAGS_NONE, "4") ;
		
		zoulou_class = c ;
		
		return MAX_ERR_NONE ;
	}

// ----------------------------------------------------------------------------------------------------------------

void *zoulou_new (t_symbol *s, long argc, t_atom *argv)
	{
		t_zoulou *x = NULL;
		
		if (x = (t_zoulou *)object_alloc (zoulou_class)) 
			{
				if (x->graf = (t_noeud **)sysmem_newptrclear (sizeof(t_noeud *) * INIT_GRAF_SIZE)) 
					{	
						t_noeud *ptr = NULL ;
						
						if (x->graf[0] = noeud_new (ptr))
							{
								srand (time(NULL)) ;
								
								x->graf_size		= INIT_GRAF_SIZE ;
								x->graf_index		= 1 ;
								x->graf_pos			= 0 ;
								x->temp				= 0 ;
								x->flag				= 0 ;
								x->slice			= 0 ;
								x->lock				= 0 ;
								x->repeat			= DEF_REPEAT ;
								x->straight			= DEF_STRAIGHT ;
								x->threshold		= DEF_THRESHOLD ;
								
								x->slice_result	= atomarray_new (0, NULL) ;
								x->get_result	= atomarray_new (0, NULL) ;
								
								x->m_clock	= clock_new ((t_object *)x, (method)zoulou_task) ;
								
								x->m_out4	= outlet_new ((t_object *)x, NULL) ;
								x->m_out3	= listout ((t_object *)x) ;
								x->m_out2	= bangout ((t_object *)x) ;
								x->m_out1	= intout ((t_object *)x) ;
								
								systhread_mutex_new (&x->m_mutex, SYSTHREAD_MUTEX_NORMAL) ;

								attr_args_process (x, argc, argv) ;
							}
						else
							{
								object_free (x) ;
								x = NULL ;  
							}
					}
				else
					{
						object_free (x) ;
						x = NULL ;   
					}
			}
				
		return x ;
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void zoulou_bang (t_zoulou *x)
	{	
		if (x->graf_index < 2)
			{
				outlet_bang (x->m_out2) ;
			}
		else
			{
				long k ;
				
				systhread_mutex_lock (x->m_mutex) ;
						
				k = graf_step (x) ;
				
				systhread_mutex_unlock (x->m_mutex) ;
				
				outlet_int (x->m_out1, k) ;
			}
	}
	
void zoulou_int (t_zoulou *x, long n)           
	{	
		if (!systhread_mutex_trylock (x->m_mutex)) 
			{	
				if (x->graf_index < TAILLE_GRAF_MAX)
					{		
						graf_add (x, n) ;
					}
				
				systhread_mutex_unlock (x->m_mutex) ;
			}

		if (x->flag)
			{
				t_atom	*av ;
				long	ac	;
				
				atomarray_getatoms (x->slice_result, &ac, &av) ;
				
				outlet_list (x->m_out3, NULL, (short)ac, av) ;
					
				x->flag = 0 ;
			}
	}
	
void zoulou_clear (t_zoulou *x)
	{
		systhread_mutex_lock (x->m_mutex) ;
				
		graf_clear (x) ;
		
		systhread_mutex_unlock (x->m_mutex) ;
	}

void zoulou_go (t_zoulou *x, long n)           
	{	
		if (!systhread_mutex_trylock (x->m_mutex)) 
			{	
				if (n >= 0 && n < (x->graf_index - 1))
					{		
						x->graf_pos = n ;
					}
				
				systhread_mutex_unlock (x->m_mutex) ;
			}
	}
	
void zoulou_get (t_zoulou *x)           
	{	
		if (!systhread_mutex_trylock (x->m_mutex)) 
			{	
				if (x->get_result)
					{
						t_noeud *nptr = NULL ;
						t_arc	*aptr1 = NULL ;
						t_arc	*aptr2 = NULL ;
						t_atom	msg[3] ;
						
						nptr = x->graf[x->graf_pos] ;
						
						atom_setsym	 (msg, my_msg_lrs) ;
						atom_setlong (msg + 1, nptr->lrs) ;
						atom_setlong (msg + 2, nptr->refer) ;
						
						atomarray_appendatoms (x->get_result, 3, msg) ;
						
						aptr1 = linklist_getindex (nptr->list_arcs, 0) ;
						while (aptr1) 
							{
								linklist_next (nptr->list_arcs, (void *)aptr1, (void *)&aptr2) ;
								
								atom_setsym	 (msg, my_msg_link) ;
								atom_setlong (msg + 1, aptr1->val) ;
								atom_setlong (msg + 2, aptr1->dest) ;
						
								atomarray_appendatoms (x->get_result, 3, msg) ;
						
								aptr1 = aptr2 ;	
							} 
						
						atom_setsym	 (msg, my_msg_straight) ;
						atom_setlong (msg + 1, nptr->straight_val) ;
						atom_setlong (msg + 2, x->graf_pos + 1) ;
						
						atomarray_appendatoms (x->get_result, 3, msg) ;
					}
					
				systhread_mutex_unlock (x->m_mutex) ;
			}
			
		if (x->get_result)
			{
				t_atom	*av ;
				long	ac ;
						
				atomarray_getatoms (x->get_result, &ac, &av) ;
				
				if (ac) 
					{
						long	i ;
						t_atom	*ptr = NULL ;
						
						for (i = 0 ; i < ac ; i += 3)
							{
								ptr = av + i ;
								
								outlet_anything (x->m_out4, atom_getsym (ptr), (short)2, ptr + 1) ;
							}
						
						atomarray_clear (x->get_result) ;
					}
			}
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void graf_add (t_zoulou *x, long n)
	{
		long	k, j, w ;
		t_noeud *nptr = NULL ;
		t_arc	*aptr = NULL ;
		long	index = x->graf_index ;
		
		if (index == x->graf_size)
			{
				t_noeud **temp = x->graf ;
			
				if (x->graf = (t_noeud **)sysmem_resizeptrclear (x->graf, sizeof(t_noeud *) * x->graf_size * 2))     
					{  
						x->graf_size = x->graf_size * 2 ;
					}
				else
					{
						x->graf = temp ; 
					}
			}
 
		if (index < x->graf_size)
			{
				t_noeud *ptr = NULL ;
				
				if (x->graf[index])
					{
						ptr = x->graf[index] ;
					}
				
				if (x->graf[index] = noeud_new (ptr)) 
					{
						nptr = x->graf[index - 1] ;
						nptr->straight_val = n ;
						
						j = nptr->refer ;
						k = index - 1 ;
						
						while (j > -1 && !graf_exist (x->graf, j, n))
							{
								nptr = x->graf[j] ;
								if (aptr = arc_new (index, n)) 
									{
										linklist_append (nptr->list_arcs, aptr) ;
									}
								k = j ;
								j = nptr->refer ;
							}
							
						if (j == -1)
							{
								w = 0 ;
							}
						else
							{
								w = graf_exist (x->graf, j, n) ;
							}
							
						nptr = x->graf[index] ;
						nptr->refer = w ;
						nptr->lrs = repeated_suffix (x->graf, k, nptr->refer) ;
						
						switch (x->slice) {
							case 0	:	break ;
							case 1	:	if (nptr->lrs < x->temp && x->temp >= x->threshold)
											{
												fill_slice_result (x, x->temp, x->graf[index - 1]->refer) ;	
											}
										break ;
							case 2	:	if (nptr->lrs >= x->threshold)
											{
												fill_slice_result (x, x->threshold, x->graf[index]->refer) ;	
											}
										break ;	
							}
	
						x->temp = nptr->lrs ;
						x->graf_index++ ;
					}
			}
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
	
t_noeud *noeud_new (t_noeud *ptr)
	{
		t_noeud *nptr = NULL ;

		if (ptr)
			{
				object_free (ptr->list_arcs) ;
				sysmem_freeptr (ptr) ;
			}
				
		if (nptr = (t_noeud *)sysmem_newptrclear (sizeof(t_noeud))) 
			{
				nptr->refer = -1 ;
				nptr->lrs = 0 ;
				nptr->straight_val = 0 ;
				nptr->list_arcs = linklist_new () ;
				linklist_flags (nptr->list_arcs, OBJ_FLAG_MEMORY) ;
			}
			
		return nptr ;
	}

t_arc *arc_new (long d, long v)
	{
		t_arc *aptr = NULL ;
		
		if (aptr = (t_arc *)sysmem_newptrclear (sizeof(t_arc)))
			{
				aptr->dest = d ;
				aptr->val = v ;
			}
			
		return aptr ;
	}
	
long graf_exist (t_noeud **p, long n, long v)
	{
		long	k = 0 ;
		t_arc	*found = NULL ;
		t_noeud *nptr = NULL ;
		
		nptr = *(p + n) ;
		
		if (nptr->straight_val == v)
			{
				k = n + 1 ;
			}
		else
			{
				linklist_findfirst (nptr->list_arcs, (void *)&found, (t_cmpfn)arc_exist, &v) ;
				if (found)
					{
						k = found->dest ;
					}
			}
			
		return k ;
	}

long arc_exist (t_arc *aptr, long *v)	
	{
		return aptr->val == *v ;
	}

long repeated_suffix (t_noeud **p, long a, long b)
	{
		if (b == 0)
			{
				return 0 ;
			}
		else
			{
				return (common_suffix (p, a, b - 1)) + 1 ;
			}
	}

long common_suffix (t_noeud **p, long a, long b)
	{
		t_noeud *nptr1 = *(p + a) ;
		t_noeud *nptr2 = *(p + b) ;
		
		if (b == nptr1->refer)
			{
				return nptr1->lrs ;
			}
		else
			{
				while (nptr2->refer != nptr1->refer)
					{
						nptr2 = *(p + nptr2->refer) ;
					}
			}
			
		return fmin (nptr1->lrs, nptr2->lrs) ;
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
	
long graf_step (t_zoulou *x) 
	{
		long	*k = NULL ;
		t_noeud *nptr = NULL ;
		long	hasard = rand() % 100 ;
		
		if (x->graf_pos == ((x->graf_index) - 1)) 
			{
				x->graf_pos = 0 ;
			}
		
		if (hasard > x->straight)
			{
				nptr = x->graf[x->graf_pos] ;
			
				if (linklist_getsize(nptr->list_arcs))
					{
						t_arc *aptr = NULL ;
						
						linklist_shuffle (nptr->list_arcs) ;
						aptr = linklist_getindex (nptr->list_arcs, 0) ;
						k = &(aptr->val) ;
						x->graf_pos = aptr->dest ;
					}
			
				if (!k)
					{
						if (nptr->lrs > x->repeat)
							{
								x->graf_pos = nptr->refer ;
							}
					}
			}
		
		if (!k)
			{
				nptr = x->graf[x->graf_pos] ;		
				k = &(nptr->straight_val) ;
				x->graf_pos++ ;
			}
			
		if (k)
			{
				return *k ;
			}
		else
			{
				return -1 ;
			}
	}

void graf_clear (t_zoulou *x)
	{
		t_noeud *nptr = NULL ;
		
		nptr = x->graf[0] ;
		
		nptr->refer = -1 ;
		nptr->lrs = 0 ;
		nptr->straight_val = 0 ;
		linklist_clear (nptr->list_arcs) ;
		
		x->graf_index = 1 ;
		x->graf_pos = 0 ;
		x->temp = 0 ;
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void fill_slice_result (t_zoulou *x, long ac, long av)
	{
		if (x->slice_result && ac <= MAX_RESULT)
			{
				long	i ;
				t_noeud *nptr = NULL ;
				
				atomarray_clear (x->slice_result) ;
				
				for (i = 0 ; i < ac ; i++)
					{
						t_atom a ;
						
						nptr = x->graf[(av - ac + i)] ;
						
						atom_setlong (&a, nptr->straight_val) ;
						atomarray_appendatom (x->slice_result, &a) ;
					}
	
				x->flag = 1 ;
			}
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void zoulou_dump (t_zoulou *x, long n)
	{
		if (!x->lock)
			{
				long	i ;
				t_noeud *nptr = NULL ;
				t_arc	*aptr1 = NULL ;
				t_arc	*aptr2 = NULL ;
				
				x->lock = 1 ;
				
				post ("Nb de noeuds : %ld / lrs : %ld / straight : %ld", x->graf_index, x->repeat, x->straight) ;
				
				for (i = MIN (n, x->graf_index) ; i < MIN (x->graf_index, n + 50) ; i++)
					{
						nptr = x->graf[i] ;
												
						post ("Noeud %ld : val %ld ref %ld lrs %ld", i, nptr->straight_val, nptr->refer, nptr->lrs) ;
							  
						aptr1 = linklist_getindex (nptr->list_arcs, 0) ;
						while (aptr1) 
							{
								linklist_next (nptr->list_arcs, (void *)aptr1, (void *)&aptr2) ;
								post ("          arc : val %ld dest %ld", aptr1->val, aptr1->dest) ;
								aptr1 = aptr2 ;	
							}  
					} 
				
				clock_fdelay (x->m_clock, 1000.) ;
			}
	}
	
void zoulou_task (t_zoulou *x)
	{
		  x->lock = 0 ;
	}

void zoulou_assist (t_zoulou *x, void *b, long m, long a, char *s)
	{
		if (m == ASSIST_INLET) 
			{ 
				sprintf (s, "(bang) (int) clear go get") ;
			} 
		else 
			{	
				switch (a) {
					case 0 :
						sprintf (s, "navigate") ;
						break ;
					case 1:
						sprintf (s, "bang if empty") ;
						break ;
					case 2 :
						sprintf (s, "repeated slices") ;
						break ;	
					case 3 :
						sprintf (s, "get") ;
						break ;
				}
			}
	}

void zoulou_free (t_zoulou *x)
	{ 
		long i ;
		t_noeud *nptr = NULL ;
		
		for (i = 0 ; i < x->graf_size ; i++) 
			{
				if (nptr = x->graf[i])
					{
						object_free (nptr->list_arcs) ;
						sysmem_freeptr (nptr) ;
					}
			}
			
		sysmem_freeptr (x->graf) ;
		
		if (x->slice_result)
			{
				object_free	(x->slice_result) ;
			}
		
		if (x->get_result)
			{
				object_free	(x->get_result) ;
			}
		
		clock_unset (x->m_clock) ;
		object_free (x->m_clock) ;
			
		systhread_mutex_free (x->m_mutex) ;
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------:x